import logging
import time

import wiremind_kubernetes
from wiremind_kubernetes import KubernetesDeploymentManager

logger = logging.getLogger(__name__)


# Side effect, but we don't really care
KubernetesDeploymentManager.SCALE_DOWN_MAX_WAIT_TIME = 30


def are_deployments_ready(
    concerned_dm: KubernetesDeploymentManager, unconcerned_dm: KubernetesDeploymentManager
) -> bool:
    return (
        concerned_dm.is_deployment_ready("concerned")
        and concerned_dm.is_deployment_ready("concerned-very-high-priority")
        and concerned_dm.is_deployment_ready("concerned-high-priority")
        and unconcerned_dm.is_deployment_ready("unconcerned")
    )


def wait_for_deployments_ready(concerned_dm: KubernetesDeploymentManager, unconcerned_dm: KubernetesDeploymentManager):
    for _ in range(1, 10):
        logger.info("Waiting for deployments to be started...")
        if not are_deployments_ready(concerned_dm, unconcerned_dm):
            logger.info("All Deployments not ready yet, waiting...")
            time.sleep(5)
        else:
            # OK!
            break
    else:
        assert are_deployments_ready(concerned_dm, unconcerned_dm)  # Last chance


def test_stop_start_all(concerned_dm, unconcerned_dm, populate_cluster, mocker):
    """
    Test that we stop/start all deployments that have an EDS in the namespace default
    and only them.
    """
    spied_scale_down_deployment = mocker.spy(wiremind_kubernetes.NamespacedKubernetesHelper, "scale_down_deployment")
    spied_are_deployments_stopped = mocker.spy(
        wiremind_kubernetes.KubernetesDeploymentManager, "_are_deployments_stopped"
    )

    concerned_dm.stop_pods()
    assert concerned_dm.is_deployment_stopped("concerned")
    assert concerned_dm.is_deployment_stopped("concerned-2")
    assert concerned_dm.is_deployment_stopped("concerned-very-high-priority")
    assert concerned_dm.is_deployment_stopped("concerned-high-priority")
    assert not unconcerned_dm.is_deployment_stopped("unconcerned")

    # Test stop order to see if we honor priority (for in-depth testing of priority, see unit tests)
    scale_down_call_list = spied_scale_down_deployment.call_args_list
    assert scale_down_call_list[0][0][1] == "concerned-very-high-priority"
    assert scale_down_call_list[-1][0][1] == "concerned-low-priority"
    mocked_wait_call_list = spied_are_deployments_stopped.call_args_list
    assert mocked_wait_call_list[0][0][1] == {"concerned-very-high-priority": 1}
    assert mocked_wait_call_list[-1][0][1] == {"concerned-low-priority": 1}

    concerned_dm.start_pods()

    wait_for_deployments_ready(concerned_dm, unconcerned_dm)

    assert not concerned_dm.is_deployment_stopped("concerned")
    assert not concerned_dm.is_deployment_stopped("concerned-2")
    assert not concerned_dm.is_deployment_stopped("concerned-high-priority")
    assert not concerned_dm.is_deployment_stopped("concerned-very-high-priority")
    assert not unconcerned_dm.is_deployment_stopped("unconcerned")



import logging
import pprint
import time

import kubernetes

from .conftest import TEST_NAMESPACE

logger = logging.getLogger(__name__)


def test_create_job(concerned_dm, create_namespace):
    """
    Test that default create job and delete job work as expected
    """
    job_name = "my-test-job"
    concerned_dm.create_job(
        concerned_dm.generate_job(
            job_name=job_name, container_image="gcr.io/google_containers/pause-amd64:3.1", labels={"foo": "bar"}
        )
    )
    for _ in range(1, 20):
        created_job = concerned_dm.client_batchv1_api.read_namespaced_job(
            concerned_dm.release_name + "-" + job_name, TEST_NAMESPACE
        )
        if created_job.status.active == 1:
            break
        else:
            logger.info("job not ready yet, waiting...")
            time.sleep(5)
    assert created_job.status.active == 1
    # The priorityclass was set to "", Kube will ignore it
    # and render the priority of the Pod to 0
    assert created_job.spec.template.spec.priority_class_name is None

    concerned_dm.delete_job(job_name=job_name)
    for _ in range(1, 10):
        try:
            concerned_dm.client_batchv1_api.read_namespaced_job(
                concerned_dm.release_name + "-" + job_name, TEST_NAMESPACE
            )
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                break
        else:
            logger.info("job is not deleted yet, waiting...")
            time.sleep(1)

    for _ in range(1, 30):
        pod_list = concerned_dm.client_corev1_api.list_namespaced_pod(
            TEST_NAMESPACE, label_selector=f"job-name={concerned_dm.release_name}-{job_name}"
        ).items
        if not pod_list:
            break
        else:
            logger.info("Related pod is not deleted yet, waiting...")
            time.sleep(1)
    pprint.pprint(pod_list)
    assert not pod_list


def test_create_job_argument(concerned_dm, create_namespace):
    """
    Test that create job with command / args works and finishes as expected
    """
    job_name = "my-test-job"
    priority_class_name = "two-thousand-and-six"
    concerned_dm.create_job(
        concerned_dm.generate_job(
            job_name=job_name,
            container_image="alpine:latest",
            command="sh",
            args=["-c", "true"],
            labels={"foo": "bar"},
            priority_class_name=priority_class_name,
        )
    )

    for _ in range(1, 20):
        created_job = concerned_dm.client_batchv1_api.read_namespaced_job(
            concerned_dm.release_name + "-" + job_name, TEST_NAMESPACE
        )
        if created_job.status.succeeded == 1:
            break
        else:
            logger.info("job not finished yet, waiting...")
            time.sleep(5)
    assert created_job.status.succeeded == 1
    # Check that the priority was set as we wanted
    assert created_job.spec.template.spec.priority_class_name == priority_class_name



import kubernetes.client.rest
import pytest

from wiremind_kubernetes.utils import retry_kubernetes_request, retry_kubernetes_request_no_ignore


def test_no_retry_required():
    counter = 0

    @retry_kubernetes_request
    def succeeds():
        nonlocal counter
        counter += 1
        return "success"

    r = succeeds()

    assert r == "success"
    assert counter == 1


def test_retries_once():
    counter = 0

    @retry_kubernetes_request
    def fails_once():
        nonlocal counter
        counter += 1
        if counter < 2:
            raise kubernetes.client.rest.ApiException("failed")
        else:
            return "success"

    r = fails_once()
    assert r == "success"
    assert counter == 2


def test_limit_is_reached():
    counter = 0

    @retry_kubernetes_request
    def always_fails():
        nonlocal counter
        counter += 1
        raise kubernetes.client.rest.ApiException("failed")

    with pytest.raises(kubernetes.client.rest.ApiException):
        always_fails()
    assert counter == 2


def test_404_correctly_handled():
    counter = 0

    @retry_kubernetes_request
    def notfound():
        nonlocal counter
        counter += 1
        raise kubernetes.client.rest.ApiException(status=404)

    r = notfound()

    assert r is None
    assert counter == 1


def test_404_correctly_ignored():
    """
    test that retry_kubernetes_request_no_ignore raises if 404 and does not retry
    """
    counter = 0

    @retry_kubernetes_request_no_ignore
    def notfound():
        nonlocal counter
        counter += 1
        raise kubernetes.client.rest.ApiException(status=404)

    with pytest.raises(kubernetes.client.rest.ApiException):
        r = notfound()
        assert r is None
    assert counter == 1



import unittest

import wiremind_kubernetes


def test_stop_pods_priority(mocker):
    """
    Test that we honor priorities when stopping workloads.

    i.e that we stop all deployments with the highest priority first, then wait for all of those to be stopped
    then continue.
    """
    mocker.patch("kubernetes.client.AppsV1Api")
    mocker.patch("kubernetes.client.CoreV1Api")
    mocker.patch("kubernetes.client.BatchV1Api")
    mocker.patch("kubernetes.client.CustomObjectsApi")

    mocker.patch(
        "wiremind_kubernetes.KubernetesDeploymentManager._get_expected_deployment_scale_dict",
        return_value={
            0: {"last": 42},
            2: {
                "first": 17,
            },
            1: {
                "second": 17,
            },
        },
    )

    mocked_stop_deployments = mocker.patch("wiremind_kubernetes.KubernetesDeploymentManager._stop_deployments")

    kdm = wiremind_kubernetes.KubernetesDeploymentManager(
        should_load_kubernetes_config=False, namespace="foo", release_name="concerned", dry_run=True
    )
    kdm.stop_pods()

    expected_calls = [
        unittest.mock.call._stop_deployments({"first": 17}),
        unittest.mock.call._stop_deployments({"second": 17}),
        unittest.mock.call._stop_deployments({"last": 42}),
    ]

    assert mocked_stop_deployments.mock_calls == expected_calls


def test_stop_deployments_correctly_wait(mocker):
    """
    Test that we wait for deployments to be stopped
    """
    mocker.patch("kubernetes.client.AppsV1Api")
    mocker.patch("kubernetes.client.CoreV1Api")
    mocker.patch("kubernetes.client.BatchV1Api")
    mocker.patch("kubernetes.client.CustomObjectsApi")

    deployment_dict = {0: {"my-pod": 42, "my-other-pod": 113}}

    mocked_are_deployments_stopped = mocker.patch(
        "wiremind_kubernetes.KubernetesDeploymentManager._are_deployments_stopped", side_effect=[False, False, True]
    )

    kdm = wiremind_kubernetes.KubernetesDeploymentManager(
        should_load_kubernetes_config=False, namespace="foo", release_name="concerned", dry_run=True
    )
    kdm._stop_deployments(deployment_dict)

    expected_calls = [
        unittest.mock.call._are_deployments_stopped(deployment_dict),
        unittest.mock.call._are_deployments_stopped(deployment_dict),
        unittest.mock.call._are_deployments_stopped(deployment_dict),
    ]

    assert mocked_are_deployments_stopped.mock_calls == expected_calls



import kubernetes.client
from typing import Any, Dict


class ClientWithArguments:
    """
    kubernetes.client.CoreApiV1 with arguments added in every function call.

    Currently add dry_run support for write functions and pretty to all.
    """

    def __init__(self, client, dry_run: bool = False):
        self.client = client()  # like kubernetes.client.CoreV1Api
        self.read_additional_arguments: Dict[str, Any] = dict(pretty=True)
        # Every request, either read or write, will have those arguments added
        self.additional_arguments: Dict[str, Any] = self.read_additional_arguments.copy()
        if dry_run:
            # Dry run, in kube API, is not true or false, but either dry_run: All or not defined.
            self.additional_arguments["dry_run"] = "All"

    def __getattr__(self, attr):
        original_attr = getattr(self.client, attr)

        if not callable(original_attr):
            return original_attr

        is_write_function: bool = False
        for name in ["create_", "delete_", "patch_", "replace_"]:
            if attr.startswith(name):
                is_write_function = True
                break

        def fn(*args, **kwargs):
            if is_write_function:
                kwargs.update(self.additional_arguments)
            else:  # A read function
                kwargs.update(self.read_additional_arguments)
            return original_attr(*args, **kwargs)

        return fn


class CoreV1ApiWithArguments(ClientWithArguments):
    def __init__(self, *args, dry_run: bool = False, **kwargs):
        super().__init__(client=kubernetes.client.CoreV1Api, dry_run=dry_run)


class AppV1ApiWithArguments(ClientWithArguments):
    def __init__(self, *args, dry_run: bool = False, **kwargs):
        super().__init__(client=kubernetes.client.AppsV1Api, dry_run=dry_run)


class BatchV1ApiWithArguments(ClientWithArguments):
    def __init__(self, *args, dry_run: bool = False, **kwargs):
        super().__init__(client=kubernetes.client.BatchV1Api, dry_run=dry_run)


class CustomObjectsApiWithArguments(ClientWithArguments):
    def __init__(self, *args, dry_run: bool = False, **kwargs):
        super().__init__(client=kubernetes.client.CustomObjectsApi, dry_run=dry_run)


class RbacAuthorizationV1ApiWithArguments(ClientWithArguments):
    def __init__(self, *args, dry_run: bool = False, **kwargs):
        super().__init__(client=kubernetes.client.RbacAuthorizationV1Api, dry_run=dry_run)



import logging
import os

import kubernetes


logger = logging.getLogger(__name__)


def _load_kubeconfig(config_file=None, context=None):
    kubernetes.config.load_kube_config(config_file=config_file, context=context)
    logger.debug("Kubernetes configuration successfully set.")


def _load_incluster_config():
    kubernetes.config.load_incluster_config()
    logger.debug("Kubernetes configuration successfully set.")


def load_kubernetes_config(use_kubeconfig=None, config_file=None, context=None):
    """
    Load kubernetes configuration in memory, either from incluster method or from kubeconfig.
    :param use_kubeconfig:
        If True: Use ~/.kube/config file to authenticate.
        If false: use kubernetes built-in incluster mechanism.
        If None, will try to load built-in incluster mechanism, then try config file.
        Defaults to None.
    :param config_file
        Name of the kube-config file.
    :param context
        Set the active context. If is set to None, current_context
        from config file will be used.
    """
    if os.environ.get("CLASSIC_K8S_CONFIG"):
        # We are in a Kind cluster for E2E test! Never use in cluster config.
        _load_kubeconfig(config_file=config_file, context=context)
        return

    if use_kubeconfig is True:
        _load_kubeconfig(config_file=config_file, context=context)
    elif use_kubeconfig is False:
        _load_incluster_config()
    elif use_kubeconfig is None:
        if os.path.exists(kubernetes.config.incluster_config.SERVICE_TOKEN_FILENAME):
            _load_incluster_config()
        else:
            _load_kubeconfig(config_file=config_file, context=context)
    else:
        logger.debug("Kubernetes configuration was not loaded!")



import logging
import os
import time

import kubernetes
import pytest

import wiremind_kubernetes
from wiremind_kubernetes.utils import run_command
from wiremind_kubernetes.tests.e2e_tests.helpers import check_not_using_wiremind_cluster

E2E_CLUSTER_MANIFESTS = "tests/e2e_tests/manifests"
absolute_path = os.path.dirname(os.path.join(os.path.abspath(wiremind_kubernetes.__file__)))
TEST_NAMESPACE = "wiremind-kube-e2e-test"


logging.getLogger("wiremind_kubernetes").setLevel(logging.DEBUG)
logger = logging.getLogger(__name__)


@pytest.fixture(scope="function")
def k8s_client_request_function(mocker):
    yield mocker.spy(kubernetes.client.api_client.ApiClient, "request")


@pytest.fixture(scope="session", autouse=True)
def setUpE2E():
    check_not_using_wiremind_cluster()


def delete_namespace():
    run_command(
        f"kubectl delete namespace {TEST_NAMESPACE} --wait --grace-period=1",
    )


@pytest.fixture
def populate_cluster():
    run_command(
        f"kubectl apply -f {absolute_path}/../../CustomResourceDefinition-expecteddeploymentscales.yaml",
    )

    try:
        run_command(
            f"kubectl create namespace {TEST_NAMESPACE}",
        )
        run_command(
            f"kubectl apply -f {absolute_path}/{E2E_CLUSTER_MANIFESTS} --namespace {TEST_NAMESPACE} --wait",
        )

        concerned_dm = wiremind_kubernetes.KubernetesDeploymentManager(
            use_kubeconfig=True,
            namespace=TEST_NAMESPACE,
            release_name="concerned",
        )

        unconcerned_dm = wiremind_kubernetes.KubernetesDeploymentManager(
            use_kubeconfig=True,
            namespace=TEST_NAMESPACE,
            release_name="unconcerned",
        )

        ready = False
        for _ in range(1, 10):
            logger.info("Waiting for deployments to be started...")
            if not concerned_dm.is_deployment_ready("concerned") or not unconcerned_dm.is_deployment_ready(
                "unconcerned"
            ):
                logger.info("All Deployments not ready yet, waiting...")
                run_command(f"kubectl get pods --namespace {TEST_NAMESPACE}")
                time.sleep(5)
            else:
                ready = True
                break
        if not ready:
            run_command(
                f"kubectl delete namespace {TEST_NAMESPACE} --wait",
            )
            raise Exception("Could not start deployments.")
    except:  # noqa E722
        delete_namespace()
        raise

    yield

    delete_namespace()


@pytest.fixture
def create_namespace():
    run_command(
        f"kubectl create namespace {TEST_NAMESPACE}",
    )

    run_command(
        f"kubectl apply -f {absolute_path}/{E2E_CLUSTER_MANIFESTS}/0_priorityclasses.yml"
        f" --namespace {TEST_NAMESPACE} --wait"
    )

    yield

    run_command(
        f"kubectl delete namespace {TEST_NAMESPACE} --wait",
    )


@pytest.fixture
def concerned_dm():
    return wiremind_kubernetes.KubernetesDeploymentManager(
        use_kubeconfig=True,
        namespace=TEST_NAMESPACE,
        release_name="concerned",
    )


@pytest.fixture
def unconcerned_dm():
    return wiremind_kubernetes.KubernetesDeploymentManager(
        use_kubeconfig=True,
        namespace=TEST_NAMESPACE,
        release_name="unconcerned",
    )



import logging
import subprocess
import sys
import urllib

logger = logging.getLogger(__name__)


def check_not_using_wiremind_cluster():
    """
    Will sys.exit(1) if kubectl current context api server is not a test cluster (like kind, minikube, etc)
    """
    logger.info("[CLUSTER-CONFIG]: Making sure the tests are not running against the main cluster...")

    api_server = subprocess.check_output(
        "kubectl config view --minify | grep server | cut -f 2- -d ':' | tr -d ' '",
        shell=True,
        universal_newlines=True,
    )
    whitelisted_api_server_hostname_list = [
        "localhost",  # minikube
        "127.0.0.1",  # kind
        "kubernetes.docker.internal",  # Docker for Mac
    ]
    hostname = urllib.parse.urlparse(api_server.lower().strip()).hostname
    if hostname not in whitelisted_api_server_hostname_list:
        logger.error("Attempted to run tests with non-test cluster %s, abort!", api_server)
        sys.exit(1)


def get_k8s_username():
    """
    Return the Kind cluster's username.
    """
    command = """kubectl config view -o json | jq '. as $o
    | ."current-context" as $current_context_name
    | $o.contexts[] | select(.name == $current_context_name) as $context
    | $o.users[] | select(.name == $context.context.user) as $user
    | $user.name'"""
    # dex-k8s-authenticator sets the user to: user={{ .Username}}-{{.ClusterName }}`
    # https://github.com/mintel/dex-k8s-authenticator/blob/master/templates/linux-mac-common.html#L101
    username = (
        subprocess.check_output(command, shell=True, universal_newlines=True).replace('"', "").strip().split("-")[0]
    )
    assert username
    return username



import subprocess

import pytest

from wiremind_kubernetes.utils import run_command, logger


def test_run_command_succeeded(mocker):
    """
    Test that running a working command works as expected.
    """
    popen_spy = mocker.spy(subprocess, "Popen")
    log_spy = mocker.spy(logger, "info")

    run_command("echo lol")

    popen_spy.assert_called_once_with(
        ["echo", "lol"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    log_spy.assert_called_with("lol")


def test_run_command_with_array_succeeded(mocker):
    """
    Test that running a working command given through array works as expected.
    """
    popen_spy = mocker.spy(subprocess, "Popen")
    log_spy = mocker.spy(logger, "info")

    run_command(["echo", "lol"])

    popen_spy.assert_called_once_with(
        ["echo", "lol"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    log_spy.assert_called_with("lol")


def test_run_command_succeeded_return_result(mocker):
    """
    Test that running a working command using return_result=True works as expected.
    """
    popen_spy = mocker.spy(subprocess, "Popen")

    stdout, stderr, returncode = run_command("echo lol", return_result=True)

    popen_spy.assert_called_once_with(
        ["echo", "lol"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    assert stdout == "lol\n"
    assert stderr is None
    assert returncode == 0


def test_run_command_succeeded_line_callback(mocker):
    """
    Test that running a working command using a custom line_callback works as expected.
    """
    result = []

    def line_callback(line):
        nonlocal result
        result.append(line)

    popen_spy = mocker.spy(subprocess, "Popen")

    run_command("echo lol", line_callback=line_callback)

    popen_spy.assert_called_once_with(
        ["echo", "lol"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    assert result == ["lol"]


def test_run_command_failed(mocker):
    """
    Test that running a failing command works as expected.
    """
    popen_spy = mocker.spy(subprocess, "Popen")
    with pytest.raises(subprocess.CalledProcessError):
        run_command("false")
    popen_spy.assert_called_once_with(
        ["false"],
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )


def test_run_command_failed_still_show_output(mocker):
    """
    Test that running a failing command still returns its output.
    """
    log_spy = mocker.spy(logger, "info")

    with pytest.raises(subprocess.CalledProcessError):
        run_command(["sh", "-c", "echo lol; false"])

    log_spy.assert_called_with("lol")


def test_run_command_failed_return_result(mocker):
    """
    Test that running a failing command using return_result=True works as expected.
    """
    stdout, stderr, returncode = run_command(["sh", "-c", "echo lol; false"], return_result=True)

    assert stdout == "lol\n"
    assert stderr is None
    assert returncode == 1


def test_run_command_failed_line_callback(mocker):
    """
    Test that running a failing command using custom line_callback works as expected.
    """
    result = []

    def line_callback(line):
        nonlocal result
        result.append(line)

    with pytest.raises(subprocess.CalledProcessError):
        run_command(["sh", "-c", "echo lol; false"], line_callback=line_callback)

    assert result == ["lol"]


def test_run_command_honors_args(mocker):
    """
    Test that run_command honors kwargs.
    """
    popen_spy = mocker.spy(subprocess, "Popen")
    log_spy = mocker.spy(logger, "info")

    run_command(["echo lol; true"], shell=True)

    popen_spy.assert_called_once_with(
        ["echo lol; true"],
        shell=True,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    log_spy.assert_called_with("lol")



class WiremindKubernetesException(Exception):
    """
    Base wiremind-kubernetes Exception.
    """

    def __init__(self, message=None):
        super().__init__()
        if message:
            self.message = message


class ExecError(WiremindKubernetesException):
    """
    An error occured while executing kubernetes command.
    """

    def __init__(self):
        super().__init__(message="An error occured while executing kubernetes command.")


class PodNotFound(WiremindKubernetesException):
    """
    A required pod was not found.
    """



import functools
import logging
import shlex
import subprocess
import time
from typing import Callable, List, Union

import kubernetes
from wiremind_kubernetes.exceptions import ExecError

logger = logging.getLogger(__name__)


def run_command(
    command: Union[List, str], return_result: bool = False, line_callback: Union[Callable, None] = None, **kw_args
):
    """
    Run command, print stdout/stderr, check that command exited correctly, return stdout/err
    """
    logger.info("Running %s", command)
    if line_callback and return_result:
        raise ValueError("line_callback and return_result parameters are mutually incompatible.")

    if not line_callback:
        line_callback = logger.info

    interpreted_command: List[str]
    if isinstance(command, str):
        interpreted_command = shlex.split(command)
    else:
        interpreted_command = command

    process = subprocess.Popen(  # type: ignore
        interpreted_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True, **kw_args
    )

    if return_result:
        out, err = process.communicate()
        return (out, err, process.returncode)

    if process.stdout:
        for line in iter(process.stdout.readline, ""):
            line_callback(line.strip())
    process.wait()

    if process.returncode:
        raise subprocess.CalledProcessError(process.returncode, command)


def retry_kubernetes_request(function):
    """
    Decorator that retries a failed Kubernetes API request if needed and ignores 404
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                logger.warning("Not found, ignoring.")
                return
            logger.error(e)
            logger.info("Retrying in 5 seconds...")
            time.sleep(5)
            return function(*args, **kwargs)
        finally:
            logger.debug("Done.")

    return wrapper


def retry_kubernetes_request_no_ignore(function):
    """
    Decorator that retries a failed Kubernetes API request if needed and do NOT ignore 404 (raise if 404)
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                raise
            logger.error(e)
            logger.info("Retrying in 5 seconds...")
            time.sleep(5)
            return function(*args, **kwargs)
        finally:
            logger.debug("Done.")

    return wrapper


def kubernetes_exec(commands, api, pod_name, namespace_name, container_name=None):
    logger.info('Connecting to "%s" pod from "%s" namespace', pod_name, namespace_name)
    resp = kubernetes.stream.stream(
        api.connect_get_namespaced_pod_exec,
        pod_name,
        namespace_name,
        command=["/bin/sh"],
        container=container_name,
        stderr=True,
        stdin=True,
        stdout=True,
        tty=False,
        _preload_content=False,
    )

    while resp.is_open():
        resp.update(timeout=1)
        if resp.peek_stdout():
            logger.info(resp.read_stdout())
        if resp.peek_stderr():
            error_message = resp.read_stderr()
            logger.error(error_message)
            if "FATAL" in error_message or "ERROR" in error_message:
                raise ExecError()
        if commands:
            c = commands.pop(0)
            logger.info("Running command: %s\n", c)
            resp.write_stdin(c + "\n")
        else:
            break



import wiremind_kubernetes


def test_is_deployment_stopped_ignores_failed(mocker):
    """
    Test that we don't consider failed (like evicted) Pods as living Pods
    """
    mocker.patch("kubernetes.client.AppsV1Api")
    mocker.patch("kubernetes.client.CoreV1Api")
    mocker.patch("kubernetes.client.BatchV1Api")
    mocker.patch("kubernetes.client.CustomObjectsApi")

    class DummyStatusObject:
        phase = "Failed"

    class DummyPodObject:
        status = DummyStatusObject()

    namespaced_kubernetes_helper = wiremind_kubernetes.NamespacedKubernetesHelper(
        should_load_kubernetes_config=False, namespace="foo"
    )

    mocker.patch(
        "wiremind_kubernetes.NamespacedKubernetesHelper._get_pods_from_deployment", return_value=[DummyPodObject()]
    )

    assert namespaced_kubernetes_helper.is_deployment_stopped("bar")



import wiremind_kubernetes.kubernetes_helper


def test_kubernetes_client_additional_arguments_core_v1_api(mocker):
    """
    Test that we add mandatory args to each function call of kubernetes client
    """
    mocked_read_namespaced_pod = mocker.patch("kubernetes.client.CoreV1Api.read_namespaced_pod")
    mocked_create_namespaced_pod = mocker.patch("kubernetes.client.CoreV1Api.create_namespaced_pod")

    kubernetes_helper = wiremind_kubernetes.kubernetes_helper.KubernetesHelper(
        dry_run=True, should_load_kubernetes_config=False
    )

    kubernetes_helper.client_corev1_api.read_namespaced_pod("foo", "bar")
    mocked_read_namespaced_pod.assert_called_once_with("foo", "bar", pretty=True)

    kubernetes_helper.client_corev1_api.create_namespaced_pod("foo", "bar")
    mocked_create_namespaced_pod.assert_called_once_with("foo", "bar", pretty=True, dry_run="All")



"""
wiremind-kubernetes
"""
from setuptools import setup, find_packages

with open("VERSION") as version_file:
    version = version_file.read().strip()

extra_require_test = [
    "mock",
    "pytest",
    "pytest-mock",
]
extra_require_mypy = [
    "mypy",
]
extra_require_dev = (
    [
        "flake8",
        "black",
        "flake8-mutable",
        "pip-tools>=3.7.0",
        "pyupgrade",
        "safety",
    ]
    + extra_require_mypy
    + extra_require_test
)


setup(
    name="wiremind-kubernetes",
    version=version,
    description="Helper for Kubernetes.",
    long_description="""This Python library is a high-level set of Kubernetes Helpers allowing either
to manage individual standard Kubernetes controllers (Deployment, StatefulSets, etc) or a logical set
of standard Kubernetes controllers through the `expecteddeploymentscales.wiremind.io` CRD (for example
allowing to scale down ALL Deployments of a Helm Release marked as such).""",
    author="Wiremind",
    author_email="dev@wiremind.io",
    url="https://github.com/wiremind/wiremind-kubernetes",
    license="LGPLv3+",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        "kubernetes>=18.0.0",
    ],
    extras_require={
        "dev": extra_require_dev,
        "mypy": extra_require_mypy,
        "test": extra_require_test,
    },
    python_requires=">=3.7.0",
    keywords=["kubernetes"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
    ],
)



import logging
import pprint
import time
from typing import Any, Dict, List, Optional, Union

import kubernetes

from wiremind_kubernetes.exceptions import PodNotFound

from .kube_config import load_kubernetes_config
from .kubernetes_client_additional_arguments import (
    AppV1ApiWithArguments,
    BatchV1ApiWithArguments,
    CoreV1ApiWithArguments,
    CustomObjectsApiWithArguments,
    RbacAuthorizationV1ApiWithArguments,
)
from .utils import retry_kubernetes_request, retry_kubernetes_request_no_ignore

logger = logging.getLogger(__name__)


class KubernetesHelper:
    """
    A simple helper for Kubernetes manipulation.
    """

    SCALE_DOWN_MAX_WAIT_TIME = 3600

    def __init__(
        self,
        use_kubeconfig: bool = False,
        dry_run: bool = False,
        should_load_kubernetes_config: bool = True,
        context: Optional[str] = None,
    ):
        """
        :param use_kubeconfig:
            Use ~/.kube/config file to authenticate.
            If false, use kubernetes built-in in_cluster mechanism.
            Defaults to False.
        :param dry_run:
            Dry run.
        :param should_load_kubernetes_config:
            if true, will call load_kubernetes_config(use_kubeconfig)
            Defaults to True.
        """
        if should_load_kubernetes_config:
            load_kubernetes_config(use_kubeconfig=use_kubeconfig, context=context)
        self.client_corev1_api: kubernetes.client.CoreV1Api = CoreV1ApiWithArguments(dry_run=dry_run)
        self.client_appsv1_api: kubernetes.client.AppsV1Api = AppV1ApiWithArguments(dry_run=dry_run)
        self.client_batchv1_api: kubernetes.client.BatchV1Api = BatchV1ApiWithArguments(dry_run=dry_run)
        self.client_custom_objects_api: kubernetes.client.CustomObjectsApi = CustomObjectsApiWithArguments(
            dry_run=dry_run
        )
        self.client_rbac_authorization_v1_api: kubernetes.client.RbacAuthorizationV1Api = (
            RbacAuthorizationV1ApiWithArguments(dry_run=dry_run)
        )

        self.dry_run: bool = dry_run


def _get_namespace_from_kube() -> str:
    return open("/var/run/secrets/kubernetes.io/serviceaccount/namespace").read()


class NamespacedKubernetesHelper(KubernetesHelper):
    """
    A simple helper for Kubernetes manipulation.
    """

    def __init__(
        self,
        use_kubeconfig: bool = False,
        namespace: Union[None, str] = None,
        dry_run: bool = False,
        should_load_kubernetes_config: bool = True,
        context: Optional[str] = None,
    ):
        """
        :param use_kubeconfig:
            Use ~/.kube/config file to authenticate.
            If false, use kubernetes built-in in_cluster mechanism.
            Defaults to False.
        :param namespace:
            Target namespace to use.
            If not defined, try to get it from kubernetes built-in serviceAccount mechanism.
        :param dry_run:
            Dry run.
        """
        super().__init__(
            use_kubeconfig=use_kubeconfig,
            dry_run=dry_run,
            should_load_kubernetes_config=should_load_kubernetes_config,
            context=context,
        )
        if namespace:
            self.namespace = namespace
        else:
            self.namespace = _get_namespace_from_kube()

    def get_deployment_scale(self, deployment_name: str):
        logger.debug("Getting deployment scale for %s", deployment_name)
        return self.client_appsv1_api.read_namespaced_deployment_scale(deployment_name, self.namespace)

    def get_statefulset_scale(self, statefulset_name: str):
        logger.debug("Getting statefulset scale for %s", statefulset_name)
        return self.client_appsv1_api.read_namespaced_stateful_set_scale(statefulset_name, self.namespace)

    def scale_down_statefulset(self, statefulset_name: str):
        body = self.get_statefulset_scale(statefulset_name)
        logger.debug("Deleting all Pods for %s", statefulset_name)
        body.spec.replicas = 0
        self.client_appsv1_api.patch_namespaced_stateful_set_scale(statefulset_name, self.namespace, body)
        logger.debug("Done deleting.")

    @retry_kubernetes_request_no_ignore
    def scale_down_deployment(self, deployment_name: str):
        body = self.get_deployment_scale(deployment_name)
        logger.debug("Deleting all Pods for %s", deployment_name)
        body.spec.replicas = 0
        self.client_appsv1_api.patch_namespaced_deployment_scale(deployment_name, self.namespace, body)
        logger.debug("Done deleting.")

    def scale_up_statefulset(self, statefulset_name: str, pod_amount: int = 1):
        body = self.get_statefulset_scale(statefulset_name)
        logger.debug("Recreating backend Pods for %s", statefulset_name)
        body.spec.replicas = pod_amount
        self.client_appsv1_api.patch_namespaced_stateful_set_scale(statefulset_name, self.namespace, body)
        logger.debug("Done recreating.")

    @retry_kubernetes_request_no_ignore
    def scale_up_deployment(self, deployment_name: str, pod_amount: int):
        body = self.get_deployment_scale(deployment_name)
        logger.debug("Recreating backend Pods for %s", deployment_name)
        body.spec.replicas = pod_amount
        self.client_appsv1_api.patch_namespaced_deployment_scale(deployment_name, self.namespace, body)
        logger.debug("Done recreating.")

    def is_statefulset_stopped(self, deployment_name: str) -> bool:
        return self.is_deployment_stopped(deployment_name, statefulset=True)

    @retry_kubernetes_request_no_ignore
    def _get_pods_from_deployment(self, deployment_name: str, statefulset: bool = False) -> List:
        if statefulset:
            logger.debug("Asking if StatefulSet %s is stopped", deployment_name)
            labels = self.client_appsv1_api.read_namespaced_stateful_set(
                deployment_name, self.namespace
            ).spec.selector.match_labels
        else:
            logger.debug("Asking if Deployment %s is stopped", deployment_name)
            labels = self.client_appsv1_api.read_namespaced_deployment(
                deployment_name, self.namespace
            ).spec.selector.match_labels

        try:
            return self.client_corev1_api.list_namespaced_pod(
                namespace=self.namespace, label_selector=",".join(["%s=%s" % kv for kv in labels.items()])
            ).items
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                return []
            else:
                raise

    def is_deployment_stopped(self, deployment_name: str, statefulset: bool = False) -> bool:
        pod_list: List = self._get_pods_from_deployment(deployment_name, statefulset)

        current_scale = 0
        for pod in pod_list:
            if pod.status.phase not in ("Failed"):
                current_scale += 1

        if current_scale > 0:
            kind = statefulset and "StatefulSet" or "Deployment"
            logger.info("%s %s has %s living replicas", deployment_name, kind, current_scale)
            return False
        return True

    def is_deployment_ready(self, deployment_name: str, statefulset: bool = False):
        if statefulset:
            status = self.client_appsv1_api.read_namespaced_stateful_set_status(deployment_name, self.namespace)
        else:
            status = self.client_appsv1_api.read_namespaced_deployment_status(deployment_name, self.namespace)
        expected_replicas = status.spec.replicas
        ready_replicas = status.status.ready_replicas
        resource_type = statefulset and "StatefulSet" or "Deployment"
        logger.debug(
            "%s %s has %s expected replicas and %s ready replicas",
            resource_type,
            deployment_name,
            expected_replicas,
            ready_replicas,
        )
        if expected_replicas == 0:
            return True
        return expected_replicas == ready_replicas

    @retry_kubernetes_request
    def getPodNameFromDeployment(self, deployment_name, namespace_name):
        """
        From a given deployment, get the first pod name
        """
        try:
            deployment = self.client_appsv1_api.read_namespaced_deployment(deployment_name, namespace_name)
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                raise PodNotFound(f"No deployment {deployment_name} was found in the namespace {namespace_name}")
            else:
                raise
        selector = ",".join(f"{key}={value}" for key, value in deployment.spec.selector.match_labels.items())
        pod_list = self.client_corev1_api.list_namespaced_pod(namespace_name, label_selector=selector).items
        if not pod_list:
            raise PodNotFound("No matching pod was found in the namespace %s" % (namespace_name))
        return pod_list[0].metadata.name


class KubernetesDeploymentManager(NamespacedKubernetesHelper):
    """
    Subclass of Kubernetes Helper allowing to scale down/up all pods that
    should be stopped/started when doing database migration/maintenance
    (alembic, dump, etc).
    The associated Deployment should define an eds.

    Usage:
    a = wiremind_kubernetes.KubernetesDeploymentManager(use_kubeconfig=True, namespace="my-namespace")
    a.stop_pods()
    do_something('wololo')
    a.start_pods()
    """

    def __init__(self, release_name: str, **kwargs):
        self.release_name = release_name
        super().__init__(**kwargs)

    @retry_kubernetes_request_no_ignore
    def _get_expected_deployment_scale_dict(self) -> Dict[int, Dict[str, int]]:
        """
        Return a dict of expected deployment scale:
        {
            0: {  # priority
                # key: Deployment name, only if it has an associated eds
                # value: expected Deployment Scale (replicas)
                "my-deployment": 3,
                "my-other-deployment": 42
            },
            1: {
                "my-third-deployment": 17,
            },
        }
        """
        logger.debug("Getting Expected Deployment Scale list")
        eds_list: List[Dict[str, Any]] = []
        release_label_keys = ["app.kubernetes.io/instance", "release"]

        for release_label_key in release_label_keys:
            logger.debug(f"Getting Expected Deployment Scale list with the" f" release label key {release_label_key}")
            try:
                eds_list.extend(
                    self.client_custom_objects_api.list_namespaced_custom_object(
                        namespace=self.namespace,
                        group="wiremind.io",
                        version="v1",
                        plural="expecteddeploymentscales",
                        label_selector=f"{release_label_key}={self.release_name}",
                    )["items"]
                )
            except kubernetes.client.rest.ApiException as e:
                if e.status != 404:
                    raise

        eds_dict: Dict[int, Dict[str, int]] = {}
        for eds in eds_list:
            deployment_name: str = eds["spec"]["deploymentName"]
            expected_scale: int = eds["spec"]["expectedScale"]
            # Note: default should be removed once we use CRD v1 with default within the CRD itself
            priority: int = eds["spec"].get("priority", 0)

            if priority not in eds_dict:
                eds_dict[priority] = {}

            eds_dict[priority][deployment_name] = expected_scale

        logger.debug("Deployments are %s", pprint.pformat(eds_dict))
        return eds_dict

    def start_pods(self):
        """
        Start all Pods that should be started
        """
        expected_deployment_scale_dict: Dict[int, Dict[str, int]] = self._get_expected_deployment_scale_dict()

        logger.info("Scaling up application Deployments...")
        if not expected_deployment_scale_dict:
            logger.info("No Deployments to scale up")
            return

        priority_dict: Dict[str, int]
        # Don't assume anything about having a priority dict within the main dict
        # So we manually test for existence
        scaled: bool = False
        for priority_dict in expected_deployment_scale_dict.values():
            name: str
            expected_scale: int
            if len(priority_dict):
                scaled = True
                for (name, expected_scale) in priority_dict.items():
                    self.scale_up_deployment(name, expected_scale)
        if scaled:
            logger.info("Done scaling up application Deployments")
        else:
            logger.info("No Deployments to scale up")

    def _are_deployments_stopped(self, deployment_dict: Dict[str, int]) -> bool:
        for deployment_name in deployment_dict:
            if not self.is_deployment_stopped(deployment_name):
                return False
        return True

    def _stop_deployments(self, deployment_dict: Dict[str, int]):
        """
        Scale down a dict (deployment_name, expected_scale) of Deployments.
        """
        for _ in range(self.SCALE_DOWN_MAX_WAIT_TIME):
            for deployment_name in deployment_dict:
                self.scale_down_deployment(deployment_name)
            if self._are_deployments_stopped(deployment_dict):
                break
            time.sleep(1)
        else:
            raise Exception("Timed out waiting for pods to be deleted: aborting.")

    def stop_pods(self):
        """
        Scale to 0 all deployments for which an ExpectedDeploymentScale links to.
        stop all deployments, then wait for actual stop, by priority (descending order):
        Example: stop all deployments with priority 1, then all deployments with priority 0
        """
        expected_deployment_scale_dict: Dict[int, Dict[str, int]] = self._get_expected_deployment_scale_dict()

        logger.info("Scaling down application Deployments...")
        if not expected_deployment_scale_dict:
            logger.info("No Deployments to scale down")
            return

        priority: int
        priorities: List[int] = sorted(expected_deployment_scale_dict, reverse=True)
        for priority in priorities:
            priority_dict: Dict[str, int] = expected_deployment_scale_dict[priority]
            if len(priority_dict):
                self._stop_deployments(priority_dict)
        logger.info("Done scaling down application Deployments.")

    def generate_job(
        self,
        job_name: str,
        container_image: str,
        labels: Dict[str, str],
        command: Union[str, None] = None,
        args: Union[List[str], None] = None,
        environment_variables: Union[Dict["str", "str"], None] = None,
        ttl_seconds_after_finished: int = 1800,
        image_pull_secrets: Union[List[kubernetes.client.V1LocalObjectReference], None] = None,
        image_pull_policy: str = "IfNotPresent",
        priority_class_name: str = "",
    ) -> kubernetes.client.V1Job:
        """
        Generate a job object.
        Note that label is mandatory since everything created in the cluster should have labels
        allowing for automatic deletion or garbage collection.
        """
        job_name = f"{self.release_name}-{job_name}"
        if environment_variables is None:
            environment_variables = {}

        job = kubernetes.client.V1Job(api_version="batch/v1", kind="Job")

        job_metadata = kubernetes.client.V1ObjectMeta(namespace=self.namespace, name=job_name)
        job_metadata.labels = labels
        job.metadata = job_metadata

        job.status = kubernetes.client.V1JobStatus()

        container = kubernetes.client.V1Container(
            name="wiremind", image=container_image, image_pull_policy=image_pull_policy
        )
        if command:
            container.command = [command]
        if args:
            container.args = args
        env_list = []
        for env_name, env_value in environment_variables.items():
            env_list.append(kubernetes.client.V1EnvVar(name=env_name, value=env_value))
        container.env = env_list

        pod_template_spec = kubernetes.client.V1PodTemplateSpec()
        pod_template_spec.spec = kubernetes.client.V1PodSpec(
            containers=[container],
            restart_policy="Never",
            image_pull_secrets=image_pull_secrets,
            priority_class_name=priority_class_name,
        )
        pod_template_spec.metadata = kubernetes.client.V1ObjectMeta(labels=labels)

        job.spec = kubernetes.client.V1JobSpec(
            ttl_seconds_after_finished=ttl_seconds_after_finished,
            template=pod_template_spec,
        )

        return job

    def create_job(self, job_body: kubernetes.client.V1Job) -> kubernetes.client.V1Job:
        try:
            return self.client_batchv1_api.create_namespaced_job(self.namespace, job_body)
        except kubernetes.client.rest.ApiException as e:
            print("Exception when calling BatchV1Api->create_namespaced_job: %s\n" % e)

    def get_job(self, job_name) -> kubernetes.client.V1Job:
        """
        Get a job, concatenating release_name and job_name as job name.
        """
        job_name = f"{self.release_name}-{job_name}"
        return self.client_batchv1_api.read_namespaced_job(job_name, self.namespace)

    def delete_job(self, job_name):
        """
        Get a job, concatenating release_name and job_name as job name.
        """
        job_name = f"{self.release_name}-{job_name}"
        body = kubernetes.client.V1DeleteOptions(propagation_policy="Background")
        return self.client_batchv1_api.delete_namespaced_job(name=job_name, namespace=self.namespace, body=body)



import os
from typing import Dict, Optional

import kubernetes
import pytest
import wiremind_kubernetes
from wiremind_kubernetes.kube_config import load_kubernetes_config


@pytest.fixture(scope="module", autouse=True)
def clean_os_environ():
    """
    Get rid of the env var "CLASSIC_K8S_CONFIG", this will be set if needed using tests parameterization
    """
    old_os_environ = os.environ.copy()
    os.environ.pop("CLASSIC_K8S_CONFIG", None)
    yield
    os.environ = old_os_environ


@pytest.mark.parametrize(
    "use_kubeconfig, config_file, context, extra_env_vars, service_token_present, should_call",
    [
        (
            True,
            None,
            "context-1",
            {},
            False,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config",
        ),
        (
            False,
            "/test/config",
            None,
            {},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config",
        ),
        (
            None,
            None,
            None,
            {},
            False,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config",
        ),
        (
            None,
            "/test/config",
            "context-2",
            {},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config",
        ),
        (
            True,
            None,
            None,
            {"CLASSIC_K8S_CONFIG": "1"},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config",
        ),
        (
            False,
            "context-1",
            "/test/config",
            {"CLASSIC_K8S_CONFIG": "1"},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config",
        ),
        (
            None,
            None,
            None,
            {"CLASSIC_K8S_CONFIG": "1"},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config",
        ),
        (
            "anything",
            None,
            None,
            {"CLASSIC_K8S_CONFIG": "1"},
            True,
            "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config",
        ),
        (
            "",
            None,
            "context-1",
            {},
            True,
            None,
        ),
        (
            "something",
            "/test/config",
            "context-1",
            {},
            False,
            None,
        ),
    ],
)
def test_load_kubernetes_config_1(
    use_kubeconfig: Optional[bool],
    config_file: Optional[str],
    context: Optional[str],
    extra_env_vars: Dict[str, str],
    service_token_present: bool,
    should_call: Optional[str],
    mocker,
):
    """
    Test that load_kubernetes_config calls the right kube loading function, when needed, with the right parameters.
    Only relevant cases are tested.
    """
    load_kube_config: str = "wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config"
    load_incluster_config: str = "wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config"

    mocker.patch(load_kube_config)
    mocker.patch(load_incluster_config)
    # merge extra_env_vars with os.environ
    mocker.patch.dict(os.environ, extra_env_vars)
    # os.path.exists is used ONLY to check for token file in wiremind_kubernetes.kube_config for now
    mocker.patch("wiremind_kubernetes.kube_config.os.path.exists", kawrgs={"side_effect": service_token_present})

    load_kubernetes_config(use_kubeconfig=use_kubeconfig, config_file=config_file, context=context)

    if should_call == load_kube_config:
        kubernetes.config.load_kube_config.assert_called_once_with(config_file=config_file, context=context)
        assert wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config.call_count == 0
    elif should_call == load_incluster_config:
        kubernetes.config.load_incluster_config.assert_called_once_with()
        assert wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config.call_count == 0
    else:
        assert wiremind_kubernetes.kube_config.kubernetes.config.load_kube_config.call_count == 0
        assert wiremind_kubernetes.kube_config.kubernetes.config.load_incluster_config.call_count == 0
